@@ -149,7 +149,7 @@ class BIMNode (GeometryNode):
 
     def get_resource(self)->o3d.geometry.TriangleMesh: 
         """Returns the mesh data in the node. \n
-        If none is present, it will search for the data on drive from path, graphPath, name or subject. 
+        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n
 
         Returns:
             o3d.geometry.TriangleMesh or None
@@ -170,6 +170,8 @@ class BIMNode (GeometryNode):
         return self.resource  
 
     def get_globalId(self):
+        """Returns the ifc globalId
+        """
         if self._globalId:
             pass 
         elif os.path.exists(self.ifcPath):
@@ -183,8 +185,8 @@ class BIMNode (GeometryNode):
         """Export the resource of the Node.\n
 
         Args:
-            directory (str, optional): directory folder to store the data.\n
-            extension (str, optional): file extension. Defaults to '.ply'.\n
+            1. directory (str, optional): directory folder to store the data.\n
+            2. extension (str, optional): file extension. Defaults to '.ply'.\n
 
         Raises:
             ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n
@@ -196,8 +198,13 @@ class BIMNode (GeometryNode):
         if self.resource is None:
             return False
         
+        #validate extension
+        if extension not in ut.MESH_EXTENSION:
+            raise ValueError('Invalid extension')
+
         # check if already exists
         if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
+            self.path=os.path.join(directory,self.get_name() + extension)
             return True
         elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:
             return True
@@ -216,11 +223,8 @@ class BIMNode (GeometryNode):
         if not os.path.exists(directory):                        
             os.mkdir(directory) 
 
-        #check extension
-        if extension in ut.MESH_EXTENSION:
-            self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)
-        else:
-            raise ValueError('Erroneous extension ')
+        self.path=os.path.join(directory,ut.get_filename(self.get_name()) + extension) #subject.toPython() replaced by get_name()
+
         #write files
         if o3d.io.write_triangle_mesh(self.path, self.resource):
             return True
@@ -229,7 +233,7 @@ class BIMNode (GeometryNode):
     def get_metadata_from_ifc_path(self) -> bool:
         """Returns the metadata from a resource. \n
 
-        Args:
+        Args (self):
             ifcPath\n
             globalId\n
 
