@@ -60,11 +60,12 @@ class ImageNode(Node):
             getResource (bool, optional= False): If True, the node will search for its physical resource on drive \n
 
         Returns:
-            A ImageNode with metadata 
+            An ImageNode with metadata 
         """  
         #private attributes 
         self._xmlPath=None
         self._xmpPath=None
+        self._orientedBoundingBox=None
         self.imageWidth = None # (int) number of pixels
         self.imageHeight = None  # (int) number of pixels
         self.focalLength35mm = None # (Float) focal length in mm     
@@ -108,7 +109,7 @@ class ImageNode(Node):
         elif (type(value) == str and value.endswith('xml') ):
             self._xmlPath=value
         else:
-            raise ValueError('self.xmlPath has invalid type, path or extension')    
+            raise ValueError('self.xmlPath has invalid type, path or extension.')    
 
   #---------------------xmpPath----------------------------
     @property
@@ -123,17 +124,43 @@ class ImageNode(Node):
         elif (type(value) == str and value.endswith('xmp') ):
             self._xmpPath=value
         else:
-            raise ValueError('self.xmpPath has invalid type, path or extension')    
+            raise ValueError('self.xmpPath has invalid type, path or extension.')    
+
+#---------------------orientedBoundingBox----------------------------
+    @property
+    def orientedBoundingBox(self): 
+        """Get the orientedBoundingBox of the Node from various inputs. \n
+
+        Args:
+            1.Open3D.geometry.OrientedBoundingBox \n
+            2.Open3D geometry\n
+
+        Returns:
+            orientedBoundingBox (o3d.geometry.OrientedBoundingBox) 
+        """
+        return self._orientedBoundingBox
+
+    @orientedBoundingBox.setter
+    def orientedBoundingBox(self,value):
+        if value is None:
+            return None
+        if 'orientedBoundingBox' in str(type(value)):
+            self._orientedBoundingBox=value
+        else:    
+            try: #geometry
+                self._orientedBoundingBox=value.get_oriented_bounding_box()
+            except:
+                raise ValueError('Input must be orientedBoundingBox (o3d.geometry.OrientedBoundingBox) or an Open3D Geometry.')
 
 #---------------------METHODS----------------------------
    
     def set_resource(self,value):
-        """Set self.resource (np.ndarray) of the Node.\n
+        """Set the resource of the Node from various inputs.\n
 
         Args:
             1. np.ndarray (OpenCV) \n
             2. PIL Image\n
-            3. Open3D Image
+            3. Open3D Image\n
 
         Raises:
             ValueError: Resource must be np.ndarray (OpenCV), PIL Image or Open3D Image.
@@ -147,8 +174,8 @@ class ImageNode(Node):
             raise ValueError('Resource must be np.ndarray (OpenCV) or PIL Image')
 
     def get_resource(self)->np.ndarray: 
-        """Returns the image data in the node. \n
-        If none is present, it will search for the data on drive from path, graphPath, name or subject. 
+        """Returns the resource (image) in the node. \n
+        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n
 
         Returns:
             np.ndarray or None
@@ -160,13 +187,15 @@ class ImageNode(Node):
         return self._resource  
 
     def get_path(self) -> str:
-        """Find the full path of the resource from this Node.\n
+        """Returns the full path of the resource from various inputs.\n
 
-        Valid inputs are:\n
-            self._graphPath and (self._name or self._subject)
+        Args:
+            1. self._graphPath \n
+            2. self._name \n
+            3. self._subject\n
 
         Returns:
-            path (str)
+            resource path (str)
         """      
         if self._path and os.path.exists(self._path):
             return self._path
@@ -208,7 +237,7 @@ class ImageNode(Node):
 
     def get_xmp_path(self)->str: 
         """Returns the xmpPath in the node. \n
-        If none is present, it will search for the data on drive from path, graphPath, name or subject. 
+        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n
 
         Returns:
             str or None
@@ -230,8 +259,8 @@ class ImageNode(Node):
         """Export the resource of the Node.\n
 
         Args:
-            directory (str, optional): directory folder to store the data.\n
-            extension (str, optional): file extension. Defaults to '.png'.\n
+            1. directory (str, optional): directory folder to store the data.\n
+            2. extension (str, optional): file extension. Defaults to '.png'.\n
 
         Raises:
             ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n
@@ -243,8 +272,13 @@ class ImageNode(Node):
         if self.resource is None:
             return False
         
+        #validate extension
+        if extension not in ut.IMG_EXTENSION:
+            raise ValueError('Invalid extension')
+
         # check if already exists
         if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
+            self.path=os.path.join(directory,self.get_name() + extension)
             return True
         elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.IMG_EXTENSION:
             return True    
@@ -263,16 +297,29 @@ class ImageNode(Node):
         if not os.path.exists(directory):                        
             os.mkdir(directory)       
 
-        #check extension
-        if extension in ut.IMG_EXTENSION:
-            self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)
-        else:
-            raise ValueError('Erroneous extension ')
+        self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)
+
         #write files
         if cv2.imwrite(self.path, self.resource):
             return True
         return False
        
+    def get_oriented_bounding_box(self)->o3d.geometry.OrientedBoundingBox:
+        """Gets the Open3D OrientedBoundingBox of the node from the conical mesh representation based on the 
+        cartesianTransform, the focal length at 35mm and a viewing range. \n
+
+        Returns:
+            o3d.geometry.orientedBoundingBox
+        """    
+        if self._orientedBoundingBox is not None:
+            pass
+        elif self._cartesianTransform is not None:
+            mesh=self.get_mesh_geometry()
+            self._orientedBoundingBox=mesh.get_oriented_bounding_box()            
+        else:
+            return None
+        return self._orientedBoundingBox
+
     def get_metadata_from_resource(self) ->bool:
         """Returns the metadata from a resource. \n
 
@@ -310,18 +357,17 @@ class ImageNode(Node):
     #     return self.orientedBoundingBox
 
     def get_mesh_geometry(self, depth:float=10, focalLength35mm:float=24)->o3d.geometry.TriangleMesh:
-        """Generate a concical mesh representation using the Image's cartesianTransform and focalLength35mm.
+        """Generate a concical mesh representation using the Image's cartesianTransform and focalLength35mm.\n
     
-        .. image:: ../docs/pics/virtual_image2.PNG
+        .. image:: ../../../docs/pics/virtual_image2.PNG
 
         Args:
-            depth (float, optional): Viewing depth of the image. Defaults to 10m.
-            focalLength35mm (float,optional): standardised focal length on 35mm film (w=36mm, h = 24mm)
+            1. depth (float, optional): Viewing depth of the image. Defaults to 10m.\n
+            2. focalLength35mm (float,optional): standardised focal length on 35mm film (w=36mm, h = 24mm)\n
 
         Returns:
-            o3d.geometry.TriangleMesh: 
+            o3d.geometry.TriangleMesh 
         """
-        
         if self.cartesianTransform is not None:
             radius=35/(focalLength35mm*2)*depth        
             mesh= o3d.geometry.TriangleMesh.create_cone(radius=radius, height=depth, resolution=20, split=1)
@@ -336,12 +382,20 @@ class ImageNode(Node):
             mesh=mesh.translate(A)
             rot=r.as_matrix()
             mesh=mesh.rotate(rot)
-
             return mesh
         else:
             return None
 
     def get_virtual_image(self, geometries: o3d.geometry, downsampling:int=2)-> o3d.cpu.pybind.geometry.Image:
+        """_summary_
+
+        Args:
+            geometries (o3d.geometry): _description_
+            downsampling (int, optional): _description_. Defaults to 2.
+
+        Returns:
+            o3d.cpu.pybind.geometry.Image: _description_
+        """
         pinholeCamera=self.get_pinhole_camera_parameters(downsampling)
         if pinholeCamera is not None:
             return gt.generate_virtual_image(geometries,pinholeCamera)
