@@ -141,7 +141,73 @@ class TestGeometryutils(unittest.TestCase):
 
 ################################## TEST FUNCTIONS ######################
 
+    def test_arrays_to_mesh_and_mesh_to_arrays(self):
+        #mesh_to_arrays
+        tuple=gt.mesh_to_arrays(self.meshPath)
+        self.assertEqual(len(tuple[0]),len(self.mesh.vertices))
+        self.assertEqual(len(tuple[1]),len(self.mesh.triangles))
+        self.assertEqual(len(tuple[2]),len(self.mesh.vertex_colors))
+        self.assertEqual(tuple[3],None)
+        self.assertEqual(tuple[4],0)
+
+        #arrays_to_mesh
+        mesh=gt.arrays_to_mesh(tuple)
+        self.assertEqual(len(mesh.vertices),len(self.mesh.vertices))
+        self.assertEqual(len(mesh.triangles),len(self.mesh.triangles))
+        self.assertEqual(len(mesh.vertex_colors),len(self.mesh.vertex_colors))
+
+    def test_arrays_to_pcd(self):
+        #pcd_to_arrays
+        tuple=gt.pcd_to_arrays(self.pcdPath1)
+        self.assertEqual(len(tuple[0]),len(self.pcd.points))
+        self.assertEqual(len(tuple[1]),len(self.pcd.colors))
+        self.assertEqual(tuple[2],None)
+        self.assertEqual(tuple[3],0)
+     
+        #arrays_to_mesh
+        pcd=gt.arrays_to_pcd(tuple)
+        self.assertEqual(len(pcd.points),len(self.pcd.points))
+        self.assertEqual(len(pcd.colors),len(self.pcd.colors))
+        self.assertEqual(len(pcd.normals),len(self.pcd.normals))
+
+    def test_create_identity_point_cloud(self):
+        # 1 geometry
+        identityPointCloud, indentityArray=gt.create_identity_point_cloud(self.slabMesh)
+        self.assertEqual(len(identityPointCloud.points),len(indentityArray))
+
+        #multiple geometries
+        list=[self.slabMesh, self.wallMesh]
+        identityPointCloud2, indentityArray2=gt.create_identity_point_cloud(list)
+        self.assertEqual(len(identityPointCloud2.points),len(indentityArray2))
+
+    def test_create_visible_point_cloud_from_meshes(self):
+        referenceMesh1= copy.deepcopy(self.slabMesh)
+        referenceMesh1.translate([1,0,0])
+        referenceMesh2= copy.deepcopy(self.slabMesh)
+        referenceMesh2.translate([0,1,0])
+        
+        # 1 geometry
+        identityPointClouds1, percentages1=gt.create_visible_point_cloud_from_meshes(geometries=self.slabMesh,
+                                                references=referenceMesh1)
+        self.assertEqual(len(identityPointClouds1),1)
+        self.assertGreater(len(identityPointClouds1[0].points),10)
+        self.assertEqual(len(percentages1),1)        
+        self.assertLess(percentages1[0],0.2)
+
+        # multiple geometries 
+        list=[self.slabMesh,self.wallMesh]
+        references=[referenceMesh1,referenceMesh2]
+        identityPointClouds2, percentages2=gt.create_visible_point_cloud_from_meshes(geometries=list,
+                                                references=references)
+        self.assertEqual(len(identityPointClouds2),2)
+        self.assertLess(len(identityPointClouds2[0].points),len(identityPointClouds1[0].points))
+        self.assertEqual(len(percentages2),2)        
+        self.assertLess(percentages2[0],percentages1[0])
+    
+        
+
     def test_cap_mesh(self):
+        print('test_cap_mesh NOT IMPLEMENTED')
         self.assertEqual(0,0)
         
     def test_box_to_mesh(self):
@@ -295,23 +361,61 @@ class TestGeometryutils(unittest.TestCase):
         image=gt.generate_virtual_image(self.mesh, pinholeCamera)
         self.assertIsInstance(image,o3d.cpu.pybind.geometry.Image)
         
-    def test_e57Path_to_pcd(self):
+    def test_e57path_to_pcd(self):
         e57=pye57.E57(self.e57Path1) 
         header = e57.get_header(0)
         pcd=gt.e57path_to_pcd(e57Path=self.e57Path1, e57Index=0) 
         self.assertEqual(len(pcd.points) , header.point_count)
 
-    def test_e57scan_to_pcd(self):
+    def test_e57_to_arrays(self):
         e57=pye57.E57(self.e57Path1) 
-        data_raw =e57.read_scan_raw(0)
-        pcd=gt.e57_to_pcd(e57)
-        self.assertEqual(len(data_raw["cartesianX"]),len(pcd.points))
+        header = e57.get_header(0)
+        tuple=gt.e57_to_arrays(self.e57Path1)
+        self.assertEqual(len(tuple),5)
+        self.assertEqual(len(tuple[0]),header.point_count)
+        self.assertEqual(len(tuple[1]),header.point_count)
+        self.assertEqual(len(tuple[2]),header.point_count)
+
+    def test_e57_to_pcd(self):
+        e57=pye57.E57(self.e57Path1) 
+        header = e57.get_header(0)
+        pcd=gt.e57_to_pcd(e57, percentage=0.5)
+        self.assertEqual(len(pcd.points),int(header.point_count*0.5))
 
         e57=pye57.E57(self.e57Path2) 
-        data_raw =e57.read_scan_raw(0)
-        pcd=gt.e57_to_pcd(e57)
-        self.assertEqual(len(data_raw["cartesianX"]),len(pcd.points))
-        
+        header = e57.get_header(1)
+        pcd=gt.e57_to_pcd(e57, percentage=0.5)
+        self.assertLess(len(pcd.points),6000000)
+
+    def test_e57path_to_pcds_multiprocessing(self):
+        e57=pye57.E57(self.e57Path1) 
+        header1 = e57.get_header(0)
+        pcds=gt.e57path_to_pcds_multiprocessing(self.e57Path1, percentage=0.5)
+        self.assertEqual(len(pcds),1)        
+        self.assertEqual(len(pcds[0].points),int(header1.point_count*0.5))
+
+        e57=pye57.E57(self.e57Path2) 
+        header1 = e57.get_header(1)
+        header2 = e57.get_header(2)
+        pcds=gt.e57path_to_pcds_multiprocessing(self.e57Path2, percentage=0.5)
+        self.assertEqual(len(pcds[0].points),int(header1.point_count*0.5))
+        self.assertEqual(len(pcds[1].points),int(header2.point_count*0.5))
+
+    def test_pcd_to_arrays(self):
+        tuple=gt.pcd_to_arrays(self.pcdPath1, percentage=0.5)
+        self.assertEqual(len(tuple),4)
+        self.assertEqual(len(tuple[0]),int(len(self.pcd.points)*0.5))
+        self.assertEqual(len(tuple[1]),int(len(self.pcd.points)*0.5))
+        self.assertEqual(tuple[2],None)
+
+    def test_mesh_to_arrays(self):
+        tuple=gt.mesh_to_arrays(self.meshPath)
+        self.assertEqual(len(tuple),5)
+        self.assertEqual(len(tuple[0]),len(self.mesh.vertices))
+        self.assertEqual(len(tuple[1]),len(self.mesh.triangles))
+        self.assertEqual(len(tuple[2]),len(self.mesh.vertex_colors))
+        self.assertEqual(tuple[3],None)
+
     def test_e57_get_colors(self):
         e57=pye57.E57(self.e57Path1) 
         raw_data = e57.read_scan_raw(0)  
@@ -319,7 +423,6 @@ class TestGeometryutils(unittest.TestCase):
         self.assertEqual(len(raw_data["cartesianX"]) , header.point_count)
 
         colors=gt.e57_get_colors(raw_data)
-        self.assertIsInstance(colors,o3d.utility.Vector3dVector)
         self.assertEqual(len(colors),len(raw_data["cartesianX"]))
         self.assertEqual(len(colors),header.point_count)
         
@@ -336,14 +439,40 @@ class TestGeometryutils(unittest.TestCase):
         self.assertIsInstance(mesh,o3d.geometry.TriangleMesh)
         self.assertGreater(len(mesh.vertices),100000)
 
+    def test_get_mesh_representation(self):
+        #mesh
+        mesh=gt.get_mesh_representation(self.mesh)
+        self.assertEqual(len(mesh.triangles),len(self.mesh.triangles))
+
+        #point cloud
+        mesh=gt.get_mesh_representation(self.pcd)
+        self.assertGreater(len(mesh.triangles),3)
+
+        #invalid input
+        self.assertRaises(ValueError,gt.get_mesh_representation,self.mesh.get_oriented_bounding_box())
+
+    def test_get_mesh_inliers(self):
+        #mesh
+        sources=[self.slabMesh,self.wallMesh]
+        indices=gt.get_mesh_inliers(sources=sources,reference=self.mesh)
+        self.assertEqual(len(indices),len(sources))
+        
+        #pcd
+        sources=[self.slabMesh,self.wallMesh]
+        indices=gt.get_mesh_inliers(sources=sources,reference=self.pcd)
+        self.assertEqual(len(indices),len(sources))
+
+        #invalid sources
+        self.assertRaises(ValueError,gt.get_mesh_inliers,self.mesh.get_oriented_bounding_box(),self.mesh.get_oriented_bounding_box())
+
     def test_expand_box(self):
-        cartesianBounds=np.array([0,3,7,9,-2,2])       
+        cartesianBounds=np.array([0,10,7,8.3,-2,2])       
         box=gt.get_oriented_bounding_box(cartesianBounds)
         #positive expansion
         expandedbox1=gt.expand_box(box,u=5,v=3,w=1)
         self.assertEqual(expandedbox1.extent[0],box.extent[0]+5)
-        self.assertEqual(expandedbox1.extent[1],box.extent[1]+1)
-        self.assertEqual(expandedbox1.extent[2],box.extent[2]+3)
+        self.assertEqual(expandedbox1.extent[1],box.extent[1]+3)
+        self.assertEqual(expandedbox1.extent[2],box.extent[2]+1)
         
         #negate expansion
         expandedbox2=gt.expand_box(box,u=-1,v=-1,w=-1)
@@ -396,7 +525,10 @@ class TestGeometryutils(unittest.TestCase):
         sources=[self.slabMesh,self.wallMesh,self.columnMesh,self.beamMesh]
         percentages=gt.determine_percentage_of_coverage(sources=sources,reference=reference)
         self.assertEqual(len(percentages), len(sources))
-        [self.assertGreater(value,0.0) for value in percentages]
+        self.assertLess(percentages[0],0.005)
+        self.assertLess(percentages[1],0.001)
+        self.assertLess(percentages[2],0.4)
+        self.assertLess(percentages[3],0.001)
 
     def test_show_geometries(self):
         # gt.show_geometries((self.bimMeshes + [self.pcd, self.mesh]), color = True)
@@ -419,10 +551,10 @@ class TestGeometryutils(unittest.TestCase):
 
         innerCrop=gt.crop_geometry_by_convex_hull(source=sourceMesh, cutters=cutters, inside = True )
         self.assertEqual(len(innerCrop),4)
-        self.assertEqual(len(innerCrop[0].vertices),12563)
-        self.assertEqual(len(innerCrop[1].vertices),814)
-        self.assertEqual(len(innerCrop[2].vertices),1036)
-        self.assertEqual(len(innerCrop[3].vertices),1266)
+        self.assertGreater(len(innerCrop[0].vertices),12000)
+        self.assertGreater(len(innerCrop[1].vertices),800)
+        self.assertGreater(len(innerCrop[2].vertices),1000)
+        self.assertGreater(len(innerCrop[3].vertices),1000)
 
         outerCrop=gt.crop_geometry_by_convex_hull(source=sourceMesh, cutters=cutters[0], inside = False ) 
         self.assertGreater(len(outerCrop[0].vertices),255000)         
